@@ -7,11 +7,11 @@ import (
 	"crypto/x509"
 	"encoding/json"
 	"encoding/pem"
+	"github.com/go-jose/go-jose/v3/jwt"
 	"strconv"
 	"strings"
 
 	"github.com/cloudentity/oauth2c/internal/oauth2"
-	"github.com/golang-jwt/jwt"
 	"github.com/grantae/certinfo"
 	"github.com/pterm/pterm"
 	"github.com/tidwall/pretty"
@@ -98,6 +98,12 @@ func LogInputData(cc oauth2.ClientConfig) {
 		{"Username", cc.Username},
 		{"Password", cc.Password},
 		{"Refresh token", cc.RefreshToken},
+		{"Signing key", cc.SigningKey},
+		{"Subject token type", cc.SubjectTokenType},
+		{"Actors token type", cc.ActorTokenType},
+		{"TLS client cert", cc.TLSCert},
+		{"TLS client key", cc.TLSKey},
+		{"TLS root CA", cc.TLSRootCA},
 	}
 
 	nonEmptyData := pterm.TableData{}
@@ -208,8 +214,9 @@ func LogRequestAndResponseln(request oauth2.Request, response interface{}) {
 
 func LogTokenPayload(response oauth2.TokenResponse) {
 	var (
-		atClaims jwt.MapClaims
-		idClaims jwt.MapClaims
+		atClaims map[string]interface{}
+		idClaims map[string]interface{}
+		err      error
 	)
 
 	if silent {
@@ -217,7 +224,7 @@ func LogTokenPayload(response oauth2.TokenResponse) {
 	}
 
 	if response.AccessToken != "" {
-		if _, _, err := parser.ParseUnverified(response.AccessToken, &atClaims); err != nil {
+		if _, atClaims, err = oauth2.UnsafeParseJWT(response.AccessToken); err != nil {
 			pterm.Error.Println(err)
 		} else {
 			pterm.Println(pterm.FgGray.Sprint("Access token:"))
@@ -226,7 +233,7 @@ func LogTokenPayload(response oauth2.TokenResponse) {
 	}
 
 	if response.IDToken != "" {
-		if _, _, err := parser.ParseUnverified(response.IDToken, &idClaims); err != nil {
+		if _, idClaims, err = oauth2.UnsafeParseJWT(response.IDToken); err != nil {
 			pterm.Error.Println(err)
 		} else {
 			pterm.Println(pterm.FgGray.Sprint("ID token:"))
@@ -256,11 +263,22 @@ func LogAuthMethod(config oauth2.ClientConfig) {
 	}
 }
 
+func LogJARM(request oauth2.Request) {
+	if silent {
+		return
+	}
+
+	if len(request.JARM) != 0 {
+		pterm.Println(pterm.FgGray.Sprint("JARM:"))
+		LogJson(request.JARM)
+	}
+}
+
 func LogAssertion(request oauth2.Request, title string, name string) {
 	var (
 		assertion = request.Form.Get(name)
-		token     *jwt.Token
-		claims    jwt.MapClaims
+		token     *jwt.JSONWebToken
+		claims    map[string]interface{}
 		err       error
 	)
 
@@ -272,13 +290,13 @@ func LogAssertion(request oauth2.Request, title string, name string) {
 		return
 	}
 
-	if token, _, err = parser.ParseUnverified(assertion, &claims); err != nil {
+	if token, claims, err = oauth2.UnsafeParseJWT(assertion); err != nil {
 		pterm.Error.Println(err)
 		return
 	}
 
-	pterm.DefaultBox.WithTitle(title).Printfln("%s = JWT-%s(payload)", name, token.Header["alg"])
-	pterm.Println("")
+	pterm.DefaultBox.WithTitle(title).Printfln("%s = JWT-%s(payload)", name, token.Headers[0].Algorithm)
+	pterm.Println()
 	pterm.Println("Payload")
 	LogJson(claims)
 	pterm.Println("")
@@ -317,20 +335,39 @@ func LogAssertion(request oauth2.Request, title string, name string) {
 		pterm.FgGray.Println(string(key))
 	}
 
-	pterm.Println("")
+	pterm.Println()
 }
 
-func LogResult(result interface{}) {
-	if !silent {
+func LogSubjectTokenAndActorToken(request oauth2.Request) {
+	var (
+		subjectToken       = request.Form.Get("subject_token")
+		actorToken         = request.Form.Get("actor_token")
+		subjectTokenClaims map[string]interface{}
+		actorTokenClaims   map[string]interface{}
+		err                error
+	)
+
+	if silent {
 		return
 	}
 
-	output, err := json.Marshal(result)
-
-	if err != nil {
-		pterm.Error.Println(err)
-		return
+	if subjectToken != "" {
+		if _, subjectTokenClaims, err = oauth2.UnsafeParseJWT(subjectToken); err != nil {
+			pterm.Error.Println(err)
+		} else {
+			pterm.Println(pterm.FgGray.Sprint("Subject token:"))
+			LogJson(subjectTokenClaims)
+		}
 	}
 
-	pterm.Println(string(output))
+	if actorToken != "" {
+		if _, actorTokenClaims, err = oauth2.UnsafeParseJWT(actorToken); err != nil {
+			pterm.Error.Println(err)
+		} else {
+			pterm.Println(pterm.FgGray.Sprint("Actor token:"))
+			LogJson(actorTokenClaims)
+		}
+	}
+
+	pterm.Println()
 }
