@@ -8,8 +8,10 @@ import (
 	"github.com/pkg/browser"
 )
 
-func AuthorizationCodeGrantFlow(clientConfig oauth2.ClientConfig, serverConfig oauth2.ServerConfig, hc *http.Client) error {
+func (c *OAuth2Cmd) AuthorizationCodeGrantFlow(clientConfig oauth2.ClientConfig, serverConfig oauth2.ServerConfig, hc *http.Client) error {
 	var (
+		parRequest       oauth2.Request
+		parResponse      oauth2.PARResponse
 		authorizeRequest oauth2.Request
 		callbackRequest  oauth2.Request
 		tokenRequest     oauth2.Request
@@ -20,15 +22,31 @@ func AuthorizationCodeGrantFlow(clientConfig oauth2.ClientConfig, serverConfig o
 
 	LogHeader("Authorization Code Flow")
 
-	// authorize endpoint
-	LogSection("Request authorization")
+	if clientConfig.PAR {
+		LogSection("Request PAR")
 
-	if authorizeRequest, codeVerifier, err = oauth2.RequestAuthorization(addr, clientConfig, serverConfig); err != nil {
-		return err
+		if parRequest, parResponse, authorizeRequest, codeVerifier, err = oauth2.RequestPAR(context.Background(), addr, clientConfig, serverConfig, hc); err != nil {
+			LogRequestAndResponseln(parRequest, err)
+			return err
+		}
+
+		LogAssertion(parRequest, "Client assertion", "client_assertion")
+		LogAuthMethod(clientConfig)
+		LogRequestAndResponse(parRequest, parResponse)
+
+		LogSection("Request authorization")
+
+		LogRequest(authorizeRequest)
+	} else {
+		LogSection("Request authorization")
+
+		if authorizeRequest, codeVerifier, err = oauth2.RequestAuthorization(addr, clientConfig, serverConfig); err != nil {
+			return err
+		}
+
+		LogRequest(authorizeRequest)
 	}
 
-	LogRequest(authorizeRequest)
-
 	if codeVerifier != "" {
 		Logln()
 		LogBox("PKCE", "code_verifier = %s\ncode_challenge = BASE64URL-ENCODE(SHA256(ASCII(code_verifier)))", codeVerifier)
@@ -45,12 +63,13 @@ func AuthorizationCodeGrantFlow(clientConfig oauth2.ClientConfig, serverConfig o
 	// callback
 	callbackStatus := LogAction("Waiting for callback. Go to the browser to authenticate...")
 
-	if callbackRequest, err = oauth2.WaitForCallback(addr); err != nil {
+	if callbackRequest, err = oauth2.WaitForCallback(clientConfig, serverConfig, addr, hc); err != nil {
 		LogRequestln(callbackRequest)
 		return err
 	}
 
 	LogRequest(callbackRequest)
+	LogJARM(callbackRequest)
 	Logln()
 
 	callbackStatus("Obtained authorization code")
@@ -76,7 +95,8 @@ func AuthorizationCodeGrantFlow(clientConfig oauth2.ClientConfig, serverConfig o
 	LogAuthMethod(clientConfig)
 	LogRequestAndResponse(tokenRequest, tokenResponse)
 	LogTokenPayloadln(tokenResponse)
-	LogResult(tokenResponse)
+
+	c.PrintResult(tokenResponse)
 
 	exchangeStatus("Exchanged authorization code for access token")
 
