@@ -3,7 +3,6 @@ package oauth2
 import (
 	"context"
 	"crypto/sha256"
-	"crypto/x509"
 	"encoding/base64"
 	"encoding/json"
 	"fmt"
@@ -16,6 +15,7 @@ import (
 	"sync"
 	"time"
 
+	"github.com/go-jose/go-jose/v3"
 	"github.com/lithammer/shortuuid/v4"
 	"github.com/pkg/errors"
 )
@@ -28,9 +28,9 @@ const (
 	PasswordGrantType          string = "password"
 	RefreshTokenGrantType      string = "refresh_token"
 	JWTBearerGrantType         string = "urn:ietf:params:oauth:grant-type:jwt-bearer"
+	TokenExchangeGrantType     string = "urn:ietf:params:oauth:grant-type:token-exchange"
+	DeviceGrantType            string = "urn:ietf:params:oauth:grant-type:device_code"
 	// CIBAGrantType              string = "urn:openid:params:grant-type:ciba"
-	// TokenExchangeGrantType     string = "urn:ietf:params:oauth:grant-type:token-exchange"
-	// DeviceGrantType            string = "urn:ietf:params:oauth:grant-type:device_code"
 )
 
 // auth methods
@@ -54,33 +54,34 @@ const CodeVerifierLength = 43
 var CodeChallengeEncoder = base64.RawURLEncoding
 
 type ClientConfig struct {
-	IssuerURL    string
-	GrantType    string
-	ClientID     string
-	ClientSecret string
-	Scopes       []string
-	AuthMethod   string
-	PKCE         bool
-	NoPKCE       bool
-	Insecure     bool
-	ResponseType []string
-	ResponseMode string
-	Username     string
-	Password     string
-	RefreshToken string
-	Assertion    string
-	SigningKey   string
-	TLSCert      string
-	TLSKey       string
-	TLSRootCA    string
+	IssuerURL        string
+	GrantType        string
+	ClientID         string
+	ClientSecret     string
+	Scopes           []string
+	AuthMethod       string
+	PKCE             bool
+	PAR              bool
+	Insecure         bool
+	ResponseType     []string
+	ResponseMode     string
+	Username         string
+	Password         string
+	RefreshToken     string
+	Assertion        string
+	SigningKey       string
+	EncryptionKey    string
+	SubjectToken     string
+	SubjectTokenType string
+	ActorToken       string
+	ActorTokenType   string
+	TLSCert          string
+	TLSKey           string
+	TLSRootCA        string
 }
 
-func RequestAuthorization(addr string, cconfig ClientConfig, sconfig ServerConfig) (r Request, codeVerifier string, err error) {
-	if r.URL, err = url.Parse(sconfig.AuthorizationEndpoint); err != nil {
-		return r, "", errors.Wrapf(err, "failed to parse authorization endpoint")
-	}
-
-	values := url.Values{
+func NewAuthorizationRequest(addr string, cconfig ClientConfig) (values url.Values, codeVerifier string, err error) {
+	values = url.Values{
 		"client_id":    {cconfig.ClientID},
 		"redirect_uri": {"http://" + addr + "/callback"},
 		"state":        {shortuuid.New()},
@@ -105,7 +106,7 @@ func RequestAuthorization(addr string, cconfig ClientConfig, sconfig ServerConfi
 		hash := sha256.New()
 
 		if _, err = hash.Write([]byte(codeVerifier)); err != nil {
-			return r, "", err
+			return values, "", err
 		}
 
 		codeChallenge := CodeChallengeEncoder.EncodeToString(hash.Sum([]byte{}))
@@ -114,20 +115,128 @@ func RequestAuthorization(addr string, cconfig ClientConfig, sconfig ServerConfi
 		values.Set("code_challenge_method", "S256")
 	}
 
+	return values, codeVerifier, nil
+}
+
+func RequestAuthorization(addr string, cconfig ClientConfig, sconfig ServerConfig) (r Request, codeVerifier string, err error) {
+	var values url.Values
+
+	if r.URL, err = url.Parse(sconfig.AuthorizationEndpoint); err != nil {
+		return r, "", errors.Wrapf(err, "failed to parse authorization endpoint")
+	}
+
+	if values, codeVerifier, err = NewAuthorizationRequest(addr, cconfig); err != nil {
+		return r, "", errors.Wrapf(err, "failed to create authorization request")
+	}
+
 	r.URL.RawQuery = values.Encode()
 	r.Method = http.MethodGet
 
 	return r, codeVerifier, nil
 }
 
-func WaitForCallback(addr string) (request Request, err error) {
+type PARResponse struct {
+	RequestURI string `json:"request_uri"`
+	ExpiresIn  int64  `json:"expires_in"`
+}
+
+func RequestPAR(
+	ctx context.Context,
+	addr string,
+	cconfig ClientConfig,
+	sconfig ServerConfig,
+	hc *http.Client,
+) (parRequest Request, parResponse PARResponse, authorizeRequest Request, codeVerifier string, err error) {
 	var (
-		srv = http.Server{Addr: addr}
-		wg  sync.WaitGroup
+		req      *http.Request
+		resp     *http.Response
+		endpoint string
+	)
+
+	// push authorization request to /par
+	if parRequest.Form, codeVerifier, err = NewAuthorizationRequest(addr, cconfig); err != nil {
+		return parRequest, parResponse, authorizeRequest, "", errors.Wrapf(err, "failed to create authorization request")
+	}
+
+	if endpoint, err = parRequest.AuthenticateClient(
+		sconfig.PushedAuthorizationRequestEndpoint,
+		sconfig.MTLsEndpointAliases.PushedAuthorizationRequestEndpoint,
+		cconfig,
+		sconfig,
+		hc,
+	); err != nil {
+		return parRequest, parResponse, authorizeRequest, "", errors.Wrapf(err, "failed to create client authentication request")
+	}
+
+	if req, err = http.NewRequestWithContext(
+		ctx,
+		http.MethodPost,
+		endpoint,
+		strings.NewReader(parRequest.Form.Encode()),
+	); err != nil {
+		return parRequest, parResponse, authorizeRequest, codeVerifier, err
+	}
+
+	if cconfig.AuthMethod == ClientSecretBasicAuthMethod {
+		req.SetBasicAuth(cconfig.ClientID, cconfig.ClientSecret)
+	}
+
+	req.Header.Add("Content-Type", "application/x-www-form-urlencoded")
+
+	parRequest.Method = req.Method
+	parRequest.Headers = req.Header
+	parRequest.URL = req.URL
+
+	if resp, err = hc.Do(req); err != nil {
+		return parRequest, parResponse, authorizeRequest, codeVerifier, err
+	}
+
+	defer resp.Body.Close()
+
+	if resp.StatusCode != http.StatusCreated {
+		return parRequest, parResponse, authorizeRequest, codeVerifier, ParseError(resp)
+	}
+
+	if err = json.NewDecoder(resp.Body).Decode(&parResponse); err != nil {
+		return parRequest, parResponse, authorizeRequest, codeVerifier, fmt.Errorf("failed to parse token response: %w", err)
+	}
+
+	// build request to /authorize
+	if authorizeRequest.URL, err = url.Parse(sconfig.AuthorizationEndpoint); err != nil {
+		return parRequest, parResponse, authorizeRequest, codeVerifier, errors.Wrapf(err, "failed to create authorization request")
+	}
+
+	values := url.Values{
+		"client_id":   {cconfig.ClientID},
+		"request_uri": {parResponse.RequestURI},
+	}
+
+	authorizeRequest.URL.RawQuery = values.Encode()
+	authorizeRequest.Method = http.MethodGet
+
+	return parRequest, parResponse, authorizeRequest, codeVerifier, nil
+}
+
+func WaitForCallback(clientConfig ClientConfig, serverConfig ServerConfig, addr string, hc *http.Client) (request Request, err error) {
+	var (
+		srv           = http.Server{Addr: addr}
+		signingKey    jose.JSONWebKey
+		encryptionKey jose.JSONWebKey
+		wg            sync.WaitGroup
 	)
 
 	wg.Add(1)
 
+	if signingKey, err = ReadKey(SigningKey, serverConfig.JWKsURI, hc); err != nil {
+		return request, errors.Wrapf(err, "failed to read signing key from %s", serverConfig.JWKsURI)
+	}
+
+	if clientConfig.EncryptionKey != "" {
+		if encryptionKey, err = ReadKey(EncryptionKey, clientConfig.EncryptionKey, hc); err != nil {
+			return request, errors.Wrapf(err, "failed to read encryption key from %s", clientConfig.EncryptionKey)
+		}
+	}
+
 	http.HandleFunc("/callback", func(w http.ResponseWriter, r *http.Request) {
 		defer func() {
 			time.AfterFunc(time.Second, func() {
@@ -138,6 +247,7 @@ func WaitForCallback(addr string) (request Request, err error) {
 		}()
 
 		if err = r.ParseForm(); err != nil {
+			log.Fatal(err)
 			return
 		}
 
@@ -145,22 +255,29 @@ func WaitForCallback(addr string) (request Request, err error) {
 		request.URL = r.URL
 		request.Form = r.PostForm
 
-		if r.URL.Query().Get("error") != "" {
+		if err = request.ParseJARM(signingKey, encryptionKey); err != nil {
+			log.Fatal(err)
+			return
+		}
+
+		w.Header().Add("Content-Type", "text/html")
+
+		if request.Get("error") != "" {
 			err = &Error{
-				ErrorCode:   r.URL.Query().Get("error"),
-				Description: r.URL.Query().Get("error_description"),
-				Hint:        r.URL.Query().Get("error_hint"),
-				TraceID:     r.URL.Query().Get("trace_id"),
+				ErrorCode:   request.Get("error"),
+				Description: request.Get("error_description"),
+				Hint:        request.Get("error_hint"),
+				TraceID:     request.Get("trace_id"),
 			}
 
 			w.WriteHeader(http.StatusBadRequest)
 
-			if _, err := w.Write([]byte(`Authorization failed. You may close this browser.`)); err != nil {
+			if _, err := w.Write([]byte(`<script>window.close()</script> Authorization failed. You may close this window.`)); err != nil {
 				log.Fatal(err)
 			}
 		} else {
 			w.WriteHeader(http.StatusOK)
-			if _, err := w.Write([]byte(`Authorization succeeded. You may close this browser.`)); err != nil {
+			if _, err := w.Write([]byte(`<script>window.close()</script> Authorization succeeded. You may close this window.`)); err != nil {
 				log.Fatal(err)
 			}
 		}
@@ -205,6 +322,7 @@ func NewTokenResponseFromForm(f url.Values) TokenResponse {
 
 type RequestTokenParams struct {
 	Code         string
+	DeviceCode   string
 	CodeVerifier string
 	RedirectURL  string
 }
@@ -217,6 +335,12 @@ func WithAuthorizationCode(code string) func(*RequestTokenParams) {
 	}
 }
 
+func WithDeviceCode(deviceCode string) func(*RequestTokenParams) {
+	return func(opts *RequestTokenParams) {
+		opts.DeviceCode = deviceCode
+	}
+}
+
 func WithCodeVerifier(codeVerifier string) func(*RequestTokenParams) {
 	return func(opts *RequestTokenParams) {
 		opts.CodeVerifier = codeVerifier
@@ -240,7 +364,7 @@ func RequestToken(
 		req      *http.Request
 		resp     *http.Response
 		params   RequestTokenParams
-		endpoint = sconfig.TokenEndpoint
+		endpoint string
 		body     []byte
 	)
 
@@ -274,47 +398,26 @@ func RequestToken(
 		}
 
 		request.Form.Set("assertion", assertion)
+	case TokenExchangeGrantType:
+		request.Form.Set("subject_token", cconfig.SubjectToken)
+		request.Form.Set("subject_token_type", cconfig.SubjectTokenType)
+
+		if cconfig.ActorToken != "" {
+			request.Form.Set("actor_token", cconfig.ActorToken)
+			request.Form.Set("actor_token_type", cconfig.ActorTokenType)
+		}
+	case DeviceGrantType:
+		request.Form.Set("device_code", params.DeviceCode)
 	}
 
-	switch cconfig.AuthMethod {
-	case NoneAuthMethod:
-		request.Form.Set("client_id", cconfig.ClientID)
-	case ClientSecretPostAuthMethod:
-		request.Form.Set("client_id", cconfig.ClientID)
-		request.Form.Set("client_secret", cconfig.ClientSecret)
-	case ClientSecretJwtAuthMethod:
-		var clientAssertion string
-
-		if clientAssertion, request.Key, err = SignJWT(
-			ClientAssertionClaims(sconfig, cconfig),
-			SecretSigner([]byte(cconfig.ClientSecret)),
-		); err != nil {
-			return request, response, err
-		}
-
-		request.Form.Set("client_assertion_type", JwtBearerClientAssertion)
-		request.Form.Set("client_assertion", clientAssertion)
-	case PrivateKeyJwtAuthMethod:
-		var clientAssertion string
-
-		if clientAssertion, request.Key, err = SignJWT(
-			ClientAssertionClaims(sconfig, cconfig),
-			JWKSigner(cconfig, hc),
-		); err != nil {
-			return request, response, err
-		}
-
-		request.Form.Set("client_assertion_type", JwtBearerClientAssertion)
-		request.Form.Set("client_assertion", clientAssertion)
-	case TLSClientAuthMethod, SelfSignedTLSAuthMethod:
-		endpoint = sconfig.MTLsEndpointAliases.TokenEndpoint
-		request.Form.Set("client_id", cconfig.ClientID)
-
-		if tr, ok := hc.Transport.(*http.Transport); ok {
-			if len(tr.TLSClientConfig.Certificates) > 0 {
-				request.Cert, _ = x509.ParseCertificate(tr.TLSClientConfig.Certificates[0].Certificate[0])
-			}
-		}
+	if endpoint, err = request.AuthenticateClient(
+		sconfig.TokenEndpoint,
+		sconfig.MTLsEndpointAliases.TokenEndpoint,
+		cconfig,
+		sconfig,
+		hc,
+	); err != nil {
+		return request, response, err
 	}
 
 	if params.RedirectURL != "" {
